'''initialize'''
from .glitch import GlitchBeautifier
from .cartoonise import CartooniseBeautifier
from .cartoongan import CartoonGanBeautifier
from .beziercurve import BezierCurveBeautifier
from .photomosaic import PhotomosaicBeautifier
from .characterize import CharacterizeBeautifier
from .douyineffect import DouyinEffectBeautifier
from .pencildrawing import PencilDrawingBeautifier
from .cartoonizeface import CartoonizeFaceBeautifier
from .fastneuralstyletransfer import FastNeuralStyleTransferBeautifier
from .geneticfitting import GeneticFittingCircleBeautifier, GeneticFittingPolygonBeautifier