'''title'''
__title__ = 'pydrawing'
'''description'''
__description__ = 'Pydrawing: Beautify your image or video.'
'''url'''
__url__ = 'https://github.com/CharlesPikachu/pydrawing'
'''version'''
__version__ = '0.1.3'
'''author'''
__author__ = 'Zhenchao Jin'
'''email'''
__email__ = 'charlesblwx@gmail.com'
'''license'''
__license__ = 'MIT License'
'''copyright'''
__copyright__ = 'Copyright 2021 Zhenchao Jin'